@@ -5,32 +5,7 @@
 # multiprocessing/pool.py
 #
 # Copyright (c) 2006-2008, R Oudkerk
-# All rights reserved.
-#
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions
-# are met:
-#
-# 1. Redistributions of source code must retain the above copyright
-#    notice, this list of conditions and the following disclaimer.
-# 2. Redistributions in binary form must reproduce the above copyright
-#    notice, this list of conditions and the following disclaimer in the
-#    documentation and/or other materials provided with the distribution.
-# 3. Neither the name of author nor the names of any contributors may be
-#    used to endorse or promote products derived from this software
-#    without specific prior written permission.
-#
-# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND
-# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
-# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
-# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
-# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
-# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
-# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-# SUCH DAMAGE.
+# Licensed to PSF under a Contributor Agreement.
 #
 from __future__ import absolute_import
 from __future__ import with_statement
@@ -75,6 +50,24 @@ else:
     from os import kill as _kill                 # noqa
 
 
+try:
+    next = next
+except NameError:
+    def next(it, *args):  # noqa
+        try:
+            return it.next()
+        except StopIteration:
+            if not args:
+                raise
+            return args[0]
+
+
+try:
+    TIMEOUT_MAX = threading.TIMEOUT_MAX
+except AttributeError:
+    TIMEOUT_MAX = 1e10
+
+
 _Semaphore = threading._Semaphore
 
 #
@@ -517,7 +510,22 @@ class TimeoutHandler(PoolThread):
         if job._timeout_callback is not None:
             job._timeout_callback(soft=False, timeout=job._timeout)
         if process:
-            process.terminate()
+            self._trywaitkill(process)
+
+    def _trywaitkill(self, worker):
+        debug('timeout: sending TERM to %s', worker._name)
+        try:
+            worker.terminate()
+        except OSError:
+            pass
+        else:
+            if worker._popen.wait(timeout=0.1):
+                return
+        debug('timeout: TERM timed-out, now sending KILL to %s', worker._name)
+        try:
+            signal.signal(worker.pid, signal.SIGKILL)
+        except OSError:
+            pass
 
     def handle_timeouts(self):
         cache = self.cache
@@ -826,7 +834,7 @@ class Pool(object):
         self.putlocks = putlocks
         self._putlock = semaphore or LaxBoundedSemaphore(self._processes)
         for i in range(processes):
-            self._create_worker_process()
+            self._create_worker_process(i)
 
         self._worker_handler = self.Supervisor(self)
         if threads:
@@ -882,7 +890,7 @@ class Pool(object):
             exitpriority=15,
             )
 
-    def _create_worker_process(self):
+    def _create_worker_process(self, i):
         sentinel = Event() if self.allow_restart else None
         w = self.Process(
             target=worker,
@@ -894,6 +902,7 @@ class Pool(object):
         self._pool.append(w)
         w.name = w.name.replace('Process', 'PoolWorker')
         w.daemon = True
+        w.index = i
         w.start()
         if self.on_process_up:
             self.on_process_up(w)
@@ -961,6 +970,12 @@ class Pool(object):
             return exitcodes.values()
         return []
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *exc_info):
+        return self.terminate()
+
     def shrink(self, n=1):
         for i, worker in enumerate(self._iterinactive()):
             self._processes -= 1
@@ -1002,9 +1017,14 @@ class Pool(object):
                     self.restart_state.step()
             except IndexError:
                 self.restart_state.step()
-            self._create_worker_process()
+            self._create_worker_process(self._avail_index())
             debug('added worker')
 
+    def _avail_index(self):
+        assert len(self._pool) < self._processes
+        indices = set(p.index for p in self._pool)
+        return next(i for i in range(self._processes) if i not in indices)
+
     def did_start_ok(self):
         return not self._join_exited_workers()
 
@@ -1243,9 +1263,9 @@ class Pool(object):
         debug('joining worker handler')
         stop_if_not_current(self._worker_handler)
         debug('joining task handler')
-        self._task_handler.stop()
+        stop_if_not_current(self._task_handler)
         debug('joining result handler')
-        self._result_handler.stop()
+        stop_if_not_current(self._result_handler)
         debug('result handler joined')
         for i, p in enumerate(self._pool):
             debug('joining worker %s/%s (%r)', i, len(self._pool), p)
@@ -1301,7 +1321,7 @@ class Pool(object):
 
         if timeout_handler is not None:
             debug('joining timeout handler')
-            timeout_handler.stop(1e100)
+            timeout_handler.stop(TIMEOUT_MAX)
 
         if pool and hasattr(pool[0], 'terminate'):
             debug('joining pool workers')
@@ -1420,6 +1440,7 @@ class MapResult(ApplyResult):
         if chunksize <= 0:
             self._number_left = 0
             self._event.set()
+            del cache[self._job]
         else:
             self._number_left = length // chunksize + bool(length % chunksize)
 
